use crate::*;
use anchor_spl::token::{transfer, Token, TokenAccount, Transfer};
use nosana_staking::StakeAccount;

#[derive(Accounts)]
pub struct Claim<'info> {
    #[account(mut)]
    pub user: Account<'info, TokenAccount>,
    #[account(mut)]
    pub vault: Account<'info, TokenAccount>,
    #[account(mut, has_one = vault @ NosanaError::InvalidVault)]
    pub reflection: Account<'info, ReflectionAccount>,
    #[account(mut, has_one = authority @ NosanaError::Unauthorized)]
    pub reward: Account<'info, RewardAccount>,
    #[account(
        has_one = authority @ NosanaError::Unauthorized,
        constraint = stake.time_unstake == 0 @ NosanaError::StakeAlreadyUnstaked,
        constraint = stake.xnos >= reward.xnos @ NosanaError::StakeDecreased,
    )]
    pub stake: Account<'info, StakeAccount>,
    #[account(mut)]
    pub authority: Signer<'info>,
    pub token_program: Program<'info, Token>,
}

pub fn handler(ctx: Context<Claim>) -> Result<()> {
    // get rewards and stats account
    let reward: &mut Account<RewardAccount> = &mut ctx.accounts.reward;
    let reflection: &mut Account<ReflectionAccount> = &mut ctx.accounts.reflection;

    // determine amount to claim
    let amount: u128 = reward.get_amount(reflection.rate);
    if amount == 0 {
        return Ok(());
    }

    // decrease the reflection pool
    reflection.remove_rewards_account(reward.reflection, reward.xnos + amount);

    // re-enter the pool with the current stake
    reward.update(
        reflection.add_rewards_account(ctx.accounts.stake.xnos, 0),
        ctx.accounts.stake.xnos,
    );

    // pay-out pending reward
    transfer(
        CpiContext::new_with_signer(
            ctx.accounts.token_program.to_account_info(),
            Transfer {
                from: ctx.accounts.vault.to_account_info(),
                to: ctx.accounts.user.to_account_info(),
                authority: ctx.accounts.vault.to_account_info(),
            },
            &[&[id::NOS_TOKEN.as_ref(), &[reflection.vault_bump]]],
        ),
        u64::try_from(amount).unwrap(),
    )
}
